#include <iostream>
#include <GL/glut.h>
#include "vmath/vmath.h"

using namespace std;

enum MatrixType {
	MTYPE_MODELVIEW		= GL_MODELVIEW_MATRIX,
	MTYPE_PROJECTION	= GL_PROJECTION_MATRIX
};

bool init();
void set_default_states();
void update_gfx();
void draw_scene();
void clean_up();

void set_glmatrix(const Matrix4x4 &mat, MatrixType mtype = MTYPE_MODELVIEW);
Matrix4x4 get_glmatrix(MatrixType mtype = MTYPE_MODELVIEW);
void update_view_matrices();
void set_light(const Vector3 &pos);
void draw_quad(const Vector3 &c1, const Vector3 &c2);
void zoom(float factor);

void key_handler(unsigned char c, int x, int y);
void mouse_handler(int x, int y);
void bn_handler(int bn, int state, int x, int y);

Vector3 cpos(0, 5, 10), lpos(20, 10, 0);
Matrix4x4 view_mat, lview_mat;
Matrix4x4 proj_mat, lproj_mat;
unsigned int ztex;

const int ztex_sz = 512;

int main(int argc, char **argv) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(800, 600);
	glutCreateWindow("Shadow mapping test");

	glutDisplayFunc(update_gfx);
	glutIdleFunc(update_gfx);
	glutKeyboardFunc(key_handler);
	glutMotionFunc(mouse_handler);
	glutMouseFunc(bn_handler);
	atexit(clean_up);

	if(!init()) return EXIT_FAILURE;

	glutMainLoop();
	return 0;
}

bool init() {
	set_default_states();

	// get the camera projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, 1.33333, 1.0, 1000.0);
	proj_mat = get_glmatrix(MTYPE_PROJECTION);

	// get the light projection matrix
	glLoadIdentity();
	gluPerspective(45.0, 1.0, 1.0, 1000.0);
	lproj_mat = get_glmatrix(MTYPE_PROJECTION);

	// create shadow buffer
	glGenTextures(1, &ztex);
	glBindTexture(GL_TEXTURE_2D, ztex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);

	// this is the only line that needs to change in order to take the zbuffer
	// instead of the color buffer later at the glCopyTexSubImage. (switch commented line).

	//glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, ztex_sz, ztex_sz, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, ztex_sz, ztex_sz, 0, GL_BGRA, GL_UNSIGNED_BYTE, 0);

	return true;
}

void set_default_states() {
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CW);

	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0);
}


// ------- THE VOODOO SWITCHES !!! ---------

#define USE_ROWS
//#define MY_VIEWMAT_INVERSION
//#define MUL_REVERSE_ORDER


void update_gfx() {
	update_view_matrices();		// update matrices from light + camera position.

	set_light(lpos);

	// ------------- Render zbuffer to the shadow map -------------

	set_glmatrix(lproj_mat, MTYPE_PROJECTION);
	set_glmatrix(lview_mat, MTYPE_MODELVIEW);	
	glViewport(0, 0, ztex_sz, ztex_sz);
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	/*
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	*/
	draw_scene();
	/*
	glEnable(GL_LIGHTING);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	*/

	glBindTexture(GL_TEXTURE_2D, ztex);
	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, ztex_sz, ztex_sz);

	// ------------ Render geometry with projected map ------------

	// prepare to render from the camera viewpoint
	glViewport(0, 0, 800, 600);
	set_glmatrix(proj_mat, MTYPE_PROJECTION);
#ifdef MY_VIEWMAT_INVERSION
	set_glmatrix(Matrix4x4::identity, MTYPE_MODELVIEW);
#else // normal way to do it ...
	set_glmatrix(view_mat, MTYPE_MODELVIEW);
#endif

	Matrix4x4 s(0.5, 0.0, 0.0, 0.5,
				0.0, 0.5, 0.0, 0.5,
				0.0, 0.0, 0.5, 0.5,
				0.0, 0.0, 0.0, 1.0);

#ifdef MY_VIEWMAT_INVERSION
	Matrix4x4 tmp = view_mat.inverse();
#else
	Matrix4x4 tmp;	// identity
#endif

#ifdef MUL_REVERSE_ORDER
	Matrix4x4 tgmat = tmp * lview_mat * lproj_mat * s;
#else // normal way to do it...
	Matrix4x4 tgmat = s * lproj_mat * lview_mat * tmp;
#endif

#ifdef USE_ROWS		// should be the normal way...
	Vector3 srow = tgmat.get_row_vector(0);
	Vector3 trow = tgmat.get_row_vector(1);
	Vector3 rrow = tgmat.get_row_vector(2);
	Vector3 qrow = tgmat.get_row_vector(3);
#else	// use columns
	Vector3 srow = tgmat.get_column_vector(0);
	Vector3 trow = tgmat.get_column_vector(1);
	Vector3 rrow = tgmat.get_column_vector(2);
	Vector3 qrow = tgmat.get_column_vector(3);
#endif
	
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGenfv(GL_S, GL_EYE_PLANE, &srow.x);
	glEnable(GL_TEXTURE_GEN_S);
	
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGenfv(GL_T, GL_EYE_PLANE, &trow.x);
	glEnable(GL_TEXTURE_GEN_T);
	
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGenfv(GL_R, GL_EYE_PLANE, &rrow.x);
	glEnable(GL_TEXTURE_GEN_R);

	glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGenfv(GL_Q, GL_EYE_PLANE, &qrow.x);
	glEnable(GL_TEXTURE_GEN_Q);
	
	// render the scene
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	set_glmatrix(view_mat, MTYPE_MODELVIEW);
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ztex);
	/*
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
	glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_INTENSITY);
	glAlphaFunc(GL_GEQUAL, 0.99);
	glEnable(GL_ALPHA_TEST);
	*/
	glDisable(GL_LIGHTING);
	draw_scene();
	glEnable(GL_LIGHTING);
	
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);
	glDisable(GL_ALPHA_TEST);

	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ztex);
	glDisable(GL_DEPTH_TEST);
	draw_quad(Vector3(0.05, 0.05), Vector3(0.3, 0.3));
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	
	glutSwapBuffers();
}

void draw_scene() {
	glutSolidTeapot(1.0);
	
	const float fy = -0.68;
	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glVertex3f(-10, fy, 10);
	glVertex3f(-10, fy, -10);
	glVertex3f(10, fy, -10);
	glVertex3f(10, fy, 10);
	glEnd();
}

void clean_up() {
}


void set_glmatrix(const Matrix4x4 &mat, MatrixType mtype) {
	glMatrixMode(mtype);
	glLoadTransposeMatrixf(mat.opengl_matrix());
}

Matrix4x4 get_glmatrix(MatrixType mtype) {
	Matrix4x4 mat;
	glGetFloatv((GLenum)mtype, (float*)&mat);
	return mat.transposed();
}

void update_view_matrices() {
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(cpos.x, cpos.y, cpos.z, 0, 0, 0, 0, 1, 0);
	view_mat = get_glmatrix(MTYPE_MODELVIEW);

	glLoadIdentity();
	gluLookAt(lpos.x, lpos.y, lpos.z, 0, 0, 0, 0, 1, 0);
	lview_mat = get_glmatrix(MTYPE_MODELVIEW);
	glPopMatrix();
}

void set_light(const Vector3 &pos) {
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	Matrix4x4 mat = view_mat;
	mat.translate(pos);
	set_glmatrix(mat, MTYPE_MODELVIEW);

	float zero[] = {0.0, 0.0, 0.0, 1.0};
	float one[] = {1.0, 1.0, 1.0, 1.0};

	glLightfv(GL_LIGHT0, GL_POSITION, zero);
	glLightfv(GL_LIGHT0, GL_AMBIENT, zero);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, one);
	glLightfv(GL_LIGHT0, GL_SPECULAR, one);
	glEnable(GL_LIGHT0);

	glPopMatrix();
}


void draw_quad(const Vector3 &c1, const Vector3 &c2) {
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0.0, 1.0, 1.0, 0.0, 0.0, 1.0);

	glBegin(GL_QUADS);
	glColor4f(1, 1, 1, 1);
	glTexCoord2f(0.0f, 1.0f);
	glVertex3f(c1.x, c1.y, -0.5);
	glTexCoord2f(1.0f, 1.0f);
	glVertex3f(c2.x, c1.y, -0.5);
	glTexCoord2f(1.0f, 0.0f);
	glVertex3f(c2.x, c2.y, -0.5);
	glTexCoord2f(0.0f, 0.0f);
	glVertex3f(c1.x, c2.y, -0.5);
	glEnd();

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}


float zoom_factor = 1.0;
void zoom(float factor) {
	cpos *= factor;
	zoom_factor *= factor;
}

void key_handler(unsigned char c, int x, int y) {
	switch(c) {
	case ',':
		lpos.x -= 0.2;
		break;

	case '.':
		lpos.x += 0.2;
		break;

	case 'a':
		zoom(0.8);
		break;

	case 'b':
		zoom(1.2);
		break;

	case 27:
		exit(0);
	}
}

int prev_x = -1, prev_y = -1;

void mouse_handler(int x, int y) {
	if(prev_x != -1) {
		float dx = (float)(x - prev_x) * 0.01;
		float dy = (float)(y - prev_y) * 0.1;
		cpos.rotate(Vector3(0, -dx, 0));
		cpos += Vector3(0, -dy * zoom_factor, 0);

		prev_x = x;
		prev_y = y;
	}
}

void bn_handler(int bn, int state, int x, int y) {
	if(bn == GLUT_RIGHT_BUTTON) {
		if(state == GLUT_DOWN) {
			prev_x = x;
			prev_y = y;
		} else {
			prev_x = prev_y = -1;
		}
	}

	if(bn == 3 && state == GLUT_DOWN) {
		zoom(0.8);
	}

	if(bn == 4 && state == GLUT_DOWN) {
		zoom(1.2);
	}
}
